//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Profile.h"
#include "StringHelp.h"
#include <shellapi.h>
#include <math.h>
#include <IniFiles.hpp>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TProfileForm *ProfileForm;
//---------------------------------------------------------------------------
__fastcall TProfileForm::TProfileForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::FormCreate(TObject *Sender)
{
    //Open Persistant Data & initialise variables
    TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

    Edit3->Text = IniFile->ReadString("GCodeProfile","TotalAngle","360");
    Edit4->Text = IniFile->ReadString("GCodeProfile","StepAngle","5");
    Edit5->Text = IniFile->ReadString("GCodeProfile","ZAxisClearance","10.0");
    Header->Text = IniFile->ReadString("GCode","Header","");
    Footer->Text = IniFile->ReadString("GCode","Footer","M2");
    OpenDialog1->InitialDir = IniFile->ReadString("GCodeProfile","InitialDir",ExtractFileDir(Application->ExeName));

    delete IniFile;

}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::FormDestroy(TObject *Sender)
{
    //Save Persistant Data
    TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

    IniFile->WriteString("GCodeProfile","TotalAngle",Edit3->Text);
    IniFile->WriteString("GCodeProfile","StepAngle",Edit4->Text);
    IniFile->WriteString("GCodeProfile","ZAxisClearance",Edit5->Text);
    IniFile->WriteString("GCode","Header",Header->Text);
    IniFile->WriteString("GCode","Footer",Footer->Text);
    IniFile->WriteString("GCodeProfile","InitialDir",OpenDialog1->InitialDir);

    delete IniFile;
    
}
//---------------------------------------------------------------------------

void __fastcall TProfileForm::OriginalGCodeButtonClick(TObject *Sender)
{
    if (OpenDialog1->Execute()) {
        Edit1->Text = OpenDialog1->FileName;
        OpenDialog1->InitialDir = ExtractFilePath(Edit1->Text);
        Edit2->Text = ChangeFileExt(Edit1->Text, " Profile.nc");
        SaveDialog1->InitialDir = ExtractFilePath(Edit1->Text);
        }

}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::NewGCodeButtonClick(TObject *Sender)
{
    if (SaveDialog1->Execute()) {
        Edit2->Text = SaveDialog1->FileName;
        SaveDialog1->InitialDir = ExtractFilePath(Edit2->Text);
        }

    
}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::ProfileButtonClick(TObject *Sender)
{
    if (!Edit1->Text.Length() || !Edit2->Text.Length()) {
        Application->MessageBox("Need two file names\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (Edit1->Text.LowerCase() == Edit2->Text.LowerCase()) {
        Application->MessageBox("Both file names the same\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (!Edit3->Text.Length()) {
        Application->MessageBox("Please enter a value for\nthe angle to be profiled\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (!Edit4->Text.Length() || ToDoubleDef(Edit4->Text,0.0) == 0.0) {
        Application->MessageBox("Please enter a value for\nthe angle subtended by a segment around an arc\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (FileExists(Edit1->Text) && !FileExists(Edit2->Text)) {
        try {
            ProfileButton->Enabled = false;
            OriginalGCodeButton->Enabled = false;
            NewGCodeButton->Enabled = false;
            HelpButton->Enabled = false;
            CloseButton->Enabled = false;
            Edit1->Enabled = false;
            Edit2->Enabled = false;
            Edit3->Enabled = false;
            Edit4->Enabled = false;
            Edit5->Enabled = false;
            Header->Enabled = false;
            Footer->Enabled = false;
            Refresh();

            TStringList *GCodeInputFile = new TStringList();
            TStringList *GCodeOutputFile = new TStringList();
            Label3->Caption = "Opening File";
            Refresh();
            GCodeInputFile->LoadFromFile(Edit1->Text);
            Label3->Caption = "Running through the file";
            Refresh();

            //do not need, or really use, all these variables, but it may be useful in the future to have them
            int counter = 0;
            AnsiString temp;
            AnsiString tempout;
            AnsiString Instruction;
            double currentX = 999.0;
            //double currentY = 999.0;
            double currentY = 0.0;
            //double currentZ = 0.0;
            double currentZ = 0.0;
            double currentA = 999.0;
            double currentB = 999.0;
            double currentC = 999.0;
            double scaleA = 1.0;
            double scaleB = 1.0;
            double scaleC = 1.0;
            double scaleX = 1.0;
            double scaleY = 1.0;
            double scaleZ = 1.0;
            double offsetA = 0.0;
            double offsetB = 0.0;
            double offsetC = 0.0;
            double offsetX = 0.0;
            double offsetY = 0.0;
            double offsetZ = 0.0;

            //use of currentMachineX allows the axes to be mapped as required while keeping the arc segment generation working
            double currentMachineX = 999.0;
            double currentFeedrate = 100.0;
            //no idea what the rapid feedrate is, so use my machine's rapid X/Y feedrate. Z is slower
            double rapidFeedrate = 2300.0;
            double currentRetractLevel = 0.0;
            double currentSpindleSpeed = 1000.0;
            TPlane currentPlane = TXYPlane;
            TSpindleDir currentSpindleDir = TCW;
            TCoolant currentCoolant = TNone;
            TUnits Units = TMetric;
            double codeno;
            double currentTime = 0.0;
            int currentTool = 0;
            bool relativeTranslation = false;
            bool relativeCentres = false;
            //max angle in radians
            double RotationAngle = ToDoubleDef(Edit3->Text,360.0);
            double StepAngle = ToDoubleDef(Edit4->Text,5.0);
            //cutter compensation
            TCutterComp cutterCompensation = TCCNone;
            TPathControl pathControl = TVelocity;

            bool unknowncode = false;
            bool firstmove = true;

            //Initialising wrapping variables
            TAxis Axis = TYAxis;
            TAxis MappedAxis = TAAxis;
            double ClearancePlane = ToDoubleDef(Edit5->Text,0.0);

            DynamicArray<TErrorType> Errors;
            DynamicArray<TGCodeDefinition> CodeList;

            //for injected lines
            int injectedcounter = 0;


            //Write file header
            GCodeOutputFile->Add("(G Code Profile Created by the Chestnut Pens GCodeProfile Program.)");
            GCodeOutputFile->Add("(Toolpath Creation Program  R.W.Williams, 2013)");
            GCodeOutputFile->Add("(Note: All dimensions are converted to mm)");
            GCodeOutputFile->Add("(Original file Z axis components are all ignored)");
            GCodeOutputFile->Add("(The clearance plane at either end of the profile = " + DP(4,ClearancePlane) + ")");
            GCodeOutputFile->Add("(  Arc Angle = " + AnsiString(RotationAngle) + "   Step Angle = " + AnsiString(StepAngle) + " )");

            //Go through the file now
            while (counter < GCodeInputFile->Count) {
                if (!(counter % 1000)) {
                    Label3->Caption = "Line " + AnsiString(counter) + " of " + AnsiString(GCodeInputFile->Count);
                    if (!(counter %10000)) {
                        Refresh();
                        } else {
                        Label3->Refresh();
                        }
                    } // end of  'if (!(counter % 1000))'
                temp = GCodeInputFile->Strings[counter];
                tempout = "";

                while (temp.Length()) {
                    temp = temp.Trim() + " ";
                    Instruction = LeftString(temp,1);

                    if (Instruction == "(") {
                        //we have a comment up to the close bracket
                        tempout = LeftString(temp,temp.Pos(")"));
                        temp = DeleteWord(temp,")").Trim();
                        //now to record this comment in the code list to regurgitate it later ad nauseam
                        CodeList.Length++;
                        CodeList[CodeList.Length - 1].Comment = true;
                        CodeList[CodeList.Length - 1].Instruction = tempout;
                        } else {
                        temp = temp.UpperCase();
                        //not a comment, so needs to be dealt with
                        //debug
                        //GCodeOutputFile->Add("(Line = " + temp + ")");
                        temp = DeleteWord(temp,Instruction);
                        codeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                        temp = DeleteWord(temp," ");

                        //now we can find the meaning of the code etc
                        if (Instruction == "F") {
                            //It's a Feed rate instruction, add it to the list
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "F";
                            CodeList[CodeList.Length - 1].Start.feedrate = currentFeedrate;
                            CodeList[CodeList.Length - 1].End.feedrate = codeno;
                            currentFeedrate = codeno;
                            if (Units == TInch) currentFeedrate = currentFeedrate * 25.4;
                            //if this feedrate is greater than the defined rapid,
                            //set this as rapid.
                            if (codeno > rapidFeedrate) rapidFeedrate = codeno;
                            }

                        if (Instruction == "G") {
                            //It's a G Code, see which one

                            switch ((int)codeno) {

                                case 1:
                                case 0:     {
                                            //Rapid Traverse G0
                                            //Normal Traverse G1
                                            //possible codes to follow: A, B, C, X, Y, Z, G53, F
                                            double feedrate;
                                            if (codeno == 0.0) {
                                                feedrate = rapidFeedrate;
                                                } else {
                                                feedrate = currentFeedrate;
                                                }
                                            double newA = currentA;
                                            double newB = currentB;
                                            double newC = currentC;
                                            double newX = currentX;
                                            double newY = currentY;
                                            double newZ = currentZ;
                                            bool G53found = false;
                                            bool feedrateset = false;

                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z" ||
                                                    LeftString(temp,1) == "F" ||
                                                    LeftString(temp,3) == "G53"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    //ChopOffFirstChars(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    double scale;
                                                    //injected lines are in mm, other lines may be in inches
                                                    if (Units == TInch && injectedcounter == 0) {
                                                        scale = 25.4;
                                                        } else {
                                                        scale = 1.0;
                                                        if (injectedcounter) injectedcounter--;
                                                        }
                                                    if (relativeTranslation) {
                                                            if (localInstruction == "A") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TAAxisInstruction;
                                                                }
                                                            if (localInstruction == "B") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TBAxisInstruction;
                                                                }
                                                            if (localInstruction == "C") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TCAxisInstruction;
                                                                }
                                                            if (localInstruction == "X") newX = currentX + localcodeno * scale;
                                                            if (localInstruction == "Y") newY = currentY + localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = currentZ + localcodeno * scale;
                                                            } else {
                                                            //Absolute movements are being used
                                                            if (localInstruction == "A") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TAAxisInstruction;
                                                                }
                                                            if (localInstruction == "B") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TBAxisInstruction;
                                                                }
                                                            if (localInstruction == "C") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TCAxisInstruction;
                                                                }
                                                            if (localInstruction == "X") newX = localcodeno * scale;
                                                            if (localInstruction == "Y") newY = localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = localcodeno * scale;
                                                            }
                                                    if (localInstruction == "F") {
                                                        feedrate = localcodeno * scale;
                                                        feedrateset = true;
                                                        }
                                                    if (localInstruction == "G") {
                                                        G53found = true;
                                                        }
                                                    temp = temp.Trim();
                                                    }//end of while statement

                                            //calculate distance to travel
                                            double dist = sqrt((newX - currentX) * (newX - currentX)
                                                            + (newY - currentY) * (newY - currentY)
                                                            + (newZ - currentZ) * (newZ - currentZ)
                                                            );
                                            double arc = sqrt((newA - currentA) * (newA - currentA)
                                                            + (newB - currentB) * (newB - currentB)
                                                            + (newC - currentC) * (newC - currentC)
                                                            );

                                            //claculate time taken to travel distance
                                            //note: feedrate is mm per min
                                            double localtime = dist * 60.0/feedrate;
                                            if (arc > dist) localtime = arc * 60.0/feedrate;
                                            /*
                                            if (firstmove) {
                                                firstmove = false;
                                                //tweaks to update positions to avoid errors
                                                currentX = newX;
                                                currentY = newY;
                                                //debug
                                                GCodeOutputFile->Add("(First Move Line = " + GCodeInputFile->Strings[counter] + ")");
                                                } else {
                                                currentTime += localtime;
                                                }
                                                */

                                            //write code to array
                                            if (currentX != newX || currentY != newY) {
                                                if (firstmove) {
                                                    firstmove = false;
                                                    //tweaks to update positions to avoid errors
                                                    currentX = newX;
                                                    currentY = newY;
                                                    //debug
                                                    //GCodeOutputFile->Add("(First Move Line = " + GCodeInputFile->Strings[counter] + ")");
                                                    } else {
                                                    currentTime += localtime;
                                                    }
                                                CodeList.Length++;
                                                CodeList[CodeList.Length - 1].Comment = false;
                                                CodeList[CodeList.Length - 1].Instruction = "G";
                                                CodeList[CodeList.Length - 1].Codeno = codeno;
                                                CodeList[CodeList.Length - 1].Start.x = currentX;
                                                CodeList[CodeList.Length - 1].Start.y = currentY;
                                                CodeList[CodeList.Length - 1].Start.z = currentZ;
                                                CodeList[CodeList.Length - 1].Start.a = currentA;
                                                CodeList[CodeList.Length - 1].Start.b = currentB;
                                                CodeList[CodeList.Length - 1].Start.c = currentC;
                                                CodeList[CodeList.Length - 1].Start.feedrate = currentFeedrate;
                                                CodeList[CodeList.Length - 1].End.x = newX;
                                                CodeList[CodeList.Length - 1].End.y = newY;
                                                CodeList[CodeList.Length - 1].End.z = newZ;
                                                CodeList[CodeList.Length - 1].End.a = newA;
                                                CodeList[CodeList.Length - 1].End.b = newB;
                                                CodeList[CodeList.Length - 1].End.c = newC;
                                                CodeList[CodeList.Length - 1].End.feedrate = feedrate;
                                                CodeList[CodeList.Length - 1].G53present = G53found;
                                                }


                                            //update position of tool & feedrate
                                            currentA = newA;
                                            currentB = newB;
                                            currentC = newC;
                                            currentX = newX;
                                            currentY = newY;
                                            currentZ = newZ;
                                            if (codeno == 1.0 && currentFeedrate != feedrate && feedrateset) {
                                                currentFeedrate = feedrate;
                                                }
                                            if (feedrate > rapidFeedrate) rapidFeedrate = feedrate;

                                            } //end of G0/G1
                                            break;
                                case 2:
                                case 3:     {
                                            TSpindleDir Arcdir;
                                            if (codeno == 3.0) {
                                                Arcdir = TCCW;
                                                } else {
                                                Arcdir = TCW;
                                                }
                                            double feedrate = currentFeedrate;
                                            double newA = currentA;
                                            double newB = currentB;
                                            double newC = currentC;
                                            double newX = currentX;
                                            double newY = currentY;
                                            double newZ = currentZ;
                                            double newI = currentX;
                                            double newJ = currentY;
                                            double newK = currentZ;
                                            bool G53found = false;
                                            bool feedrateset = false;
                                            //debuggery
                                            //GCodeOutputFile->Add("(" + temp + ")");
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z" ||
                                                    LeftString(temp,1) == "F" ||
                                                    LeftString(temp,1) == "I" ||
                                                    LeftString(temp,1) == "J" ||
                                                    LeftString(temp,1) == "K" ||
                                                    LeftString(temp,3) == "G53"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    //ChopOffFirstChars(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    double scale;
                                                    if (Units == TInch) {
                                                        scale = 25.4;
                                                        } else {
                                                        scale = 1.0;
                                                        }
                                                    if (relativeTranslation) {
                                                            if (localInstruction == "A") newA = currentA + localcodeno * scale;
                                                            if (localInstruction == "B") newB = currentB + localcodeno * scale;
                                                            if (localInstruction == "C") newC = currentC + localcodeno * scale;
                                                            if (localInstruction == "X") newX = currentX + localcodeno * scale;
                                                            if (localInstruction == "Y") newY = currentY + localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = currentZ + localcodeno * scale;
                                                            } else {
                                                            if (localInstruction == "A") newA = localcodeno * scale;
                                                            if (localInstruction == "B") newB = localcodeno * scale;
                                                            if (localInstruction == "C") newC = localcodeno * scale;
                                                            if (localInstruction == "X") newX = localcodeno * scale;
                                                            if (localInstruction == "Y") newY = localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = localcodeno * scale;
                                                            }
                                                    if (relativeCentres) {
                                                        if (localInstruction == "I") newI += localcodeno * scale;
                                                        if (localInstruction == "J") newJ += localcodeno * scale;
                                                        if (localInstruction == "K") newK += localcodeno * scale;
                                                        } else {
                                                        if (localInstruction == "I") newI = localcodeno * scale;
                                                        if (localInstruction == "J") newJ = localcodeno * scale;
                                                        if (localInstruction == "K") newK = localcodeno * scale;
                                                        }
                                                    if (localInstruction == "F") {
                                                        feedrate = localcodeno * scale;
                                                        feedrateset = true;
                                                        }
                                                    if (localInstruction == "G") {
                                                        G53found = true;
                                                        }
                                                    temp = temp.Trim();
                                                    }//end of while statement
                                            //OK, all arc information has been read in.

                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = currentX;
                                            CodeList[CodeList.Length - 1].Start.y = currentY;
                                            CodeList[CodeList.Length - 1].Start.z = currentZ;
                                            CodeList[CodeList.Length - 1].Start.a = currentA;
                                            CodeList[CodeList.Length - 1].Start.b = currentB;
                                            CodeList[CodeList.Length - 1].Start.c = currentC;
                                            CodeList[CodeList.Length - 1].Start.feedrate = currentFeedrate;
                                            CodeList[CodeList.Length - 1].End.x = newX;
                                            CodeList[CodeList.Length - 1].End.y = newY;
                                            CodeList[CodeList.Length - 1].End.z = newZ;
                                            CodeList[CodeList.Length - 1].End.a = newA;
                                            CodeList[CodeList.Length - 1].End.b = newB;
                                            CodeList[CodeList.Length - 1].End.c = newC;
                                            CodeList[CodeList.Length - 1].End.feedrate = feedrate;
                                            CodeList[CodeList.Length - 1].Centre.i = newI;
                                            CodeList[CodeList.Length - 1].Centre.j = newJ;
                                            CodeList[CodeList.Length - 1].Centre.k = newK;
                                            CodeList[CodeList.Length - 1].G53present = G53found;

                                            if (feedrateset) currentFeedrate = feedrate;

                                            //update position of tool & feedrate
                                            currentA = newA;
                                            currentB = newB;
                                            currentC = newC;
                                            currentX = newX;
                                            currentY = newY;
                                            currentZ = newZ;

                                            } //end of G2/G3
                                            break;
                                case  4:    {
                                            //Pause
                                            double pause;
                                            bool pausefound = false;
                                            while (temp.Length() && (LeftString(temp,1) == "P" )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    //ChopOffFirstChars(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    temp = temp.Trim();
                                                    pause = localcodeno;
                                                    pausefound = true;
                                                    }//end of while statement
                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            if (pausefound) {
                                                CodeList[CodeList.Length - 1].Start.x = pause;
                                                } else {
                                                CodeList[CodeList.Length - 1].Start.x = 0.0;
                                                }
                                            }
                                            break;
                                case 17:    {
                                            //XY plane
                                            currentPlane = TXYPlane;
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 18:    {
                                            //XZ plane
                                            currentPlane = TXZPlane;
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TXZPresent;
                                            //throw an error rather than record it
                                            }
                                            break;
                                case 19:    {
                                            //YZ plane
                                            currentPlane = TYZPlane;
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TYZPresent;
                                            }
                                            break;
                                case 70:
                                case 20:    {
                                            //inch units. This is the 21st Century not the 19th Century, I can't be doing with inches, convert to metric and be done with the silly units.
                                            Units = TInch;
                                            }
                                            break;
                                case 71:
                                case 21:    {
                                            //metric units.
                                            Units = TMetric;
                                            //units in the profram are metric anyway, so only re-inforce it.
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 40:    {
                                            //Cutter Compensation off.
                                            cutterCompensation = TCCNone;
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 42:
                                case 41:    {
                                            //Cutter Compensation left or right.
                                            while (temp.Length() && (LeftString(temp,1) == "D" ||
                                                    LeftString(temp,1) == "P"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    temp = DeleteWord(temp," ");
                                                    }
                                            }
                                            break;
                                case 44:
                                case 43:    {
                                            //Tool Length Compensation
                                            //don't need to store data with these
                                            //but do need to swallow the unwanted associated data
                                            while (temp.Length() && (LeftString(temp,1) == "H" )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    temp = DeleteWord(temp," ");
                                                    }
                                            }
                                            break;
                                case 49:    {
                                            //tool length compensation off
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 50:    {
                                            //scale factors turned to 1.0 in each axis
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = scaleX;
                                            CodeList[CodeList.Length - 1].Start.y = scaleY;
                                            CodeList[CodeList.Length - 1].Start.z = scaleZ;
                                            CodeList[CodeList.Length - 1].Start.a = scaleA;
                                            CodeList[CodeList.Length - 1].Start.b = scaleB;
                                            CodeList[CodeList.Length - 1].Start.c = scaleC;
                                            CodeList[CodeList.Length - 1].End.x = 1.0;
                                            CodeList[CodeList.Length - 1].End.y = 1.0;
                                            CodeList[CodeList.Length - 1].End.z = 1.0;
                                            CodeList[CodeList.Length - 1].End.a = 1.0;
                                            CodeList[CodeList.Length - 1].End.b = 1.0;
                                            CodeList[CodeList.Length - 1].End.c = 1.0;
                                            scaleX = 1.0;
                                            scaleY = 1.0;
                                            scaleZ = 1.0;
                                            scaleA = 1.0;
                                            scaleB = 1.0;
                                            scaleC = 1.0;
                                            }
                                            break;
                                case 51:    {
                                            //scale factors specified for some or all axes
                                            //not going to keep a record of them in the program, as it doesn't seem very useful
                                            tempout = tempout + "G51 ";
                                            double newScaleA = scaleA;
                                            double newScaleB = scaleB;
                                            double newScaleC = scaleC;
                                            double newScaleX = scaleX;
                                            double newScaleY = scaleY;
                                            double newScaleZ = scaleZ;
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "A") newScaleA = localcodeno;
                                                    if (localInstruction == "B") newScaleB = localcodeno;
                                                    if (localInstruction == "C") newScaleC = localcodeno;
                                                    if (localInstruction == "X") newScaleX = localcodeno;
                                                    if (localInstruction == "Y") newScaleY = localcodeno;
                                                    if (localInstruction == "Z") newScaleZ = localcodeno;
                                                    }
                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = scaleX;
                                            CodeList[CodeList.Length - 1].Start.y = scaleY;
                                            CodeList[CodeList.Length - 1].Start.z = scaleZ;
                                            CodeList[CodeList.Length - 1].Start.a = scaleA;
                                            CodeList[CodeList.Length - 1].Start.b = scaleB;
                                            CodeList[CodeList.Length - 1].Start.c = scaleC;
                                            CodeList[CodeList.Length - 1].End.x = newScaleX;
                                            CodeList[CodeList.Length - 1].End.y = newScaleY;
                                            CodeList[CodeList.Length - 1].End.z = newScaleZ;
                                            CodeList[CodeList.Length - 1].End.a = newScaleA;
                                            CodeList[CodeList.Length - 1].End.b = newScaleB;
                                            CodeList[CodeList.Length - 1].End.c = newScaleC;

                                            scaleX = newScaleX;
                                            scaleY = newScaleY;
                                            scaleZ = newScaleZ;
                                            scaleA = newScaleA;
                                            scaleB = newScaleB;
                                            scaleC = newScaleC;
                                            }
                                            break;
                                case 92:
                                case 52:    {
                                            //offsets specified for some or all axes
                                            tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            double newOffsetA = offsetA;
                                            double newOffsetB = offsetB;
                                            double newOffsetC = offsetC;
                                            double newOffsetX = offsetX;
                                            double newOffsetY = offsetY;
                                            double newOffsetZ = offsetZ;
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "A") {
                                                        newOffsetA = localcodeno;
                                                        }
                                                    if (localInstruction == "B") {
                                                        newOffsetB = localcodeno;
                                                        }
                                                    if (localInstruction == "C") {
                                                        newOffsetC = localcodeno;
                                                        }
                                                    if (localInstruction == "X") {
                                                        newOffsetX = localcodeno;
                                                        }
                                                    if (localInstruction == "Y") {
                                                        newOffsetY = localcodeno;
                                                        }
                                                    if (localInstruction == "Z") {
                                                        newOffsetZ = localcodeno;
                                                        }
                                                    }
                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = offsetX;
                                            CodeList[CodeList.Length - 1].Start.y = offsetY;
                                            CodeList[CodeList.Length - 1].Start.z = offsetZ;
                                            CodeList[CodeList.Length - 1].Start.a = offsetA;
                                            CodeList[CodeList.Length - 1].Start.b = offsetB;
                                            CodeList[CodeList.Length - 1].Start.c = offsetC;
                                            CodeList[CodeList.Length - 1].End.x = newOffsetX;
                                            CodeList[CodeList.Length - 1].End.y = newOffsetY;
                                            CodeList[CodeList.Length - 1].End.z = newOffsetZ;
                                            CodeList[CodeList.Length - 1].End.a = newOffsetA;
                                            CodeList[CodeList.Length - 1].End.b = newOffsetB;
                                            CodeList[CodeList.Length - 1].End.c = newOffsetC;

                                            offsetX = newOffsetX;
                                            offsetY = newOffsetY;
                                            offsetZ = newOffsetZ;
                                            offsetA = newOffsetA;
                                            offsetB = newOffsetB;
                                            offsetC = newOffsetC;
                                            }
                                            break;
                                case 58:
                                case 57:
                                case 56:
                                case 55:
                                case 54:    {
                                            //work offset
                                            //don't do anything
                                            }
                                            break;
                                case 59:    {
                                            //work offset with P qualifier
                                            //swallow code, don't do anything
                                            while (temp.Length() && (LeftString(temp,1) == "P" )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    temp = DeleteWord(temp," ");
                                                    }
                                            }
                                            break;
                                case 61:    {
                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) pathControl;
                                            CodeList[CodeList.Length - 1].End.x = (double) TExact;
                                            //path control mode = exact stop
                                            pathControl = TExact;
                                            }
                                case 64:    {
                                            //write code to array
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) pathControl;
                                            CodeList[CodeList.Length - 1].End.x = (double) TVelocity;
                                            //path control mode = velociy stop
                                            pathControl = TVelocity;
                                            }
                                case 68:    {
                                            //rotate co-ordinate system
                                            //causes error
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TCoOrdRotation;
                                            //swallow next data
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "R" ||
                                                    LeftString(temp,1) == "I"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    temp = DeleteWord(temp," ");
                                                    }
                                            }
                                            break;
                                case 69:    {
                                            //cancel co-ordinate rotation
                                            }
                                case 90:    {
                                            //Absolute Distance. Now may be G90 or G90.1
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;

                                            if (codeno == 90.1) {
                                                //circle centres done absolute
                                                if (relativeCentres) {
                                                    CodeList[CodeList.Length - 1].Start.x = 91.1;
                                                    } else {
                                                    CodeList[CodeList.Length - 1].Start.x = 90.1;
                                                    }
                                                relativeCentres = false;
                                                } else {
                                                //translations done absolute
                                                if (relativeTranslation) {
                                                    CodeList[CodeList.Length - 1].Start.x = 91.0;
                                                    } else {
                                                    CodeList[CodeList.Length - 1].Start.x = 90.0;
                                                    }
                                                relativeTranslation = false;
                                                }
                                            }
                                            break;
                                case 91:    {
                                            //Absolute Distance. Now may be G91 or G91.1
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "G";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            if (codeno == 91.1) {
                                                //circle centres done relative
                                                if (relativeCentres) {
                                                    CodeList[CodeList.Length - 1].Start.x = 91.1;
                                                    } else {
                                                    CodeList[CodeList.Length - 1].Start.x = 90.1;
                                                    }
                                                relativeCentres = true;
                                                } else {
                                                //translations done relative
                                                if (relativeTranslation) {
                                                    CodeList[CodeList.Length - 1].Start.x = 91.0;
                                                    } else {
                                                    CodeList[CodeList.Length - 1].Start.x = 90.0;
                                                    }
                                                relativeTranslation = false;
                                                }
                                            }
                                            break;
                                case 94:
                                case 93:    {
                                            //feed rate definition, 93 = per time, 94 = per min
                                            //swallow code, can't be bothered to do anything with these
                                            }
                                            break;
                                case 99:
                                case 98:    {
                                            //canned cycle return, different meanings allocated dependant on codeno
                                            //swallow code, can't be bothered to do anything with these
                                            }
                                            break;
                                default:    break;
                                }
                            }  //end of G codes

                        if (Instruction == "M") {
                            //it's an M Code
                            switch ((int)codeno) {
                                case 2:
                                case 1:
                                case 0:     {
                                            //Program Stop
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 3:     {
                                            //Rotate Spindle CW
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentSpindleDir;
                                            currentSpindleDir = TCW;
                                            }
                                            break;
                                case 4:     {
                                            //Rotate Spindle CCW
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentSpindleDir;
                                            currentSpindleDir = TCCW;
                                            }
                                            break;
                                case 5:     {
                                            //Stop Spindle
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentSpindleDir;
                                            currentSpindleDir = TStop;
                                            }
                                            break;
                                case 6:     {
                                            //Change tool
                                            //swallow code, can't be bothered to do anything with these
                                            }
                                            break;
                                case 7:     {
                                            //Coolant Mist
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentCoolant;
                                            currentCoolant = TMist;
                                            }
                                            break;
                                case 8:     {
                                            //Coolant Flood
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentCoolant;
                                            currentCoolant = TFlood;
                                            }
                                            break;
                                case 9:     {
                                            //Coolant None
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            CodeList[CodeList.Length - 1].Start.x = (double) currentCoolant;
                                            currentCoolant = TNone;
                                            }
                                            break;
                                case 30:    {
                                            //Program Stop & Rewind
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 47:    {
                                            //Program Stop & Rewind & restart from line 1
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 48:    {
                                            //Enable feed override
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 49:    {
                                            //Disable feed override
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                case 60:    {
                                            //Program Stop
                                            CodeList.Length++;
                                            CodeList[CodeList.Length - 1].Comment = false;
                                            CodeList[CodeList.Length - 1].Instruction = "M";
                                            CodeList[CodeList.Length - 1].Codeno = codeno;
                                            }
                                            break;
                                default:    {
                                            //un recognised code found
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TUnknownMCode;
                                            }
                                }
                            } //end of M codes  
                        if (Instruction == "N") {
                            //it's a Line number
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "N";
                            CodeList[CodeList.Length - 1].Codeno = codeno;
                            }
                        if (Instruction == "O") {
                            //it's a subroutine label number
                            //swallow subroutine labels
                            }
                        if (Instruction == "P") {
                            //it's a pause, in milliseconds
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "P";
                            CodeList[CodeList.Length - 1].Codeno = codeno;
                            currentTime += codeno/1000.0;
                            }
                        if (Instruction == "R") {
                            //Canned Cycle retract level
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "R";
                            CodeList[CodeList.Length - 1].Codeno = codeno;
                            CodeList[CodeList.Length - 1].Start.x = currentRetractLevel;
                            currentRetractLevel = codeno;
                            }
                        if (Instruction == "S") {
                            //it's spindle speed, rpm
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "S";
                            CodeList[CodeList.Length - 1].Codeno = codeno;
                            CodeList[CodeList.Length - 1].Start.x = currentSpindleSpeed;
                            currentSpindleSpeed = codeno;
                            }
                        if (Instruction == "T") {
                            //it's a tool selection number
                            CodeList.Length++;
                            CodeList[CodeList.Length - 1].Comment = false;
                            CodeList[CodeList.Length - 1].Instruction = "T";
                            CodeList[CodeList.Length - 1].Codeno = codeno;
                            CodeList[CodeList.Length - 1].Start.x = (double) currentTool;
                            currentTool = (int)codeno;
                            }

                        } //end of 'if (Instruction == "(")'
                    } //end of 'while (temp.Length())'

                //if ((tempout.Trim()).Length()) GCodeOutputFile->Add(tempout.Trim());
                //debug
                //GCodeOutputFile->Add("(X=" + DP(4,currentX) + " Y=" + DP(4,currentY) + " Z=" + DP(4,currentZ) + " A=" + DP(4,currentA) + " B=" + DP(4,currentB) + " C=" + DP(4,currentC) + ")");
                counter ++;
                } //end of 'while (counter < filelength)'

            //Having harvested the data, now is the time to use it
            if (Header->Text.Length()) {
                GCodeOutputFile->Add("(Header)");
                GCodeOutputFile->Add(Header->Text);
                }
            GCodeOutputFile->Add("(XZ Plane, Metric, Cutter Comp. off, Tool len. comp. off, Vel mode, Abs. Trans., Abs. Arc Crs)");
            GCodeOutputFile->Add("G18 G21 G40 G49 G64 G90 G90.1");
            GCodeOutputFile->Add("(Go to clearance plane)");
            GCodeOutputFile->Add("G0 Z" + DP(4,ClearancePlane));
            counter = 0;

            /*
            while (counter < CodeList.Length && CodeList[counter].Instruction != "G" &&
                    //(CodeList[counter].Codeno != 0.0 || CodeList[counter].Codeno != 1.0 || CodeList[counter].Codeno != 2.0 || CodeList[counter].Codeno != 3.0)
                    ((CodeList[counter].Codeno > 3.0)  || (CodeList[counter].Codeno < 3.1 && CodeList[counter].Start.y != CodeList[counter].End.y) || CodeList[counter].Comment)
                    ) {
                counter++;
                }
            */
            bool found;
            while (counter < CodeList.Length && !found) {
                if ((!CodeList[counter].Comment) && (CodeList[counter].Instruction == "G") && (CodeList[counter].Codeno < 4.0) && (CodeList[counter].Start.y != CodeList[counter].End.y)) {
                    found = true;
                    counter--;
                    }
                counter++;
                }
            /*
            //debug
            for (int i = 0; i < CodeList.Length; i++) {
                GCodeOutputFile->Add("(Item " + AnsiString(i) + "  Comment = " + (CodeList[i].Comment ? "true" : "false") + "  Inst = " + CodeList[i].Instruction + "  CodeNo = " + AnsiString(CodeList[i].Codeno) + ")");
                GCodeOutputFile->Add("(    Sx " + DP(4,CodeList[i].Start.x) + " Sy " + DP(4,CodeList[i].Start.y) +  "  Sz " + DP(4,CodeList[i].Start.z) +  "  Sf " + DP(4,CodeList[i].Start.feedrate) + ")");
                GCodeOutputFile->Add("(    Ex " + DP(4,CodeList[i].End.x) + " Ey " + DP(4,CodeList[i].End.y) +  "  Ez " + DP(4,CodeList[i].End.z) +  "  Ef " + DP(4,CodeList[i].End.feedrate) + ")");
                }
            GCodeOutputFile->Add("(Counter = " + AnsiString(counter) + ")");
            //end debug
            */
            if (counter < CodeList.Length) {
                //ok have established that there is actually some movement to the machine head, so let's do something with it
                GCodeOutputFile->Add("(Go to start point)");
                GCodeOutputFile->Add("G0 X" + DP(4,CodeList[counter].Start.x) + " Y0");
                double currentangle = 0.0;
                double startZ = CodeList[counter].Start.y;
                double endZ;
                while (currentangle < RotationAngle) {
                    //run through the codelist forwards
                    GCodeOutputFile->Add("(Forwards Pass at angle " + DP(4,currentangle) + " )");
                    GCodeOutputFile->Add("G0 A" + DP(4,currentangle));
                    GCodeOutputFile->Add("G1 Z" + DP(4,startZ));
                    for (int i = 0; i < CodeList.Length; i++) {
                        if (CodeList[i].Comment) {
                            GCodeOutputFile->Add(CodeList[i].Instruction);
                            } else {
                            if (CodeList[i].Instruction == "F") {
                                GCodeOutputFile->Add("F" + DP(4,CodeList[i].End.feedrate));
                                } //End of F code
                            if (CodeList[i].Instruction == "G") {
                                switch ((int) CodeList[i].Codeno) {
                                    case 0: {
                                            if (CodeList[i].End.x != CodeList[i].Start.x || CodeList[i].End.y != CodeList[i].Start.y) {
                                                tempout = "G0";
                                                if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                                if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                                GCodeOutputFile->Add(tempout);
                                                endZ = CodeList[i].End.y;
                                                }
                                            }
                                            break;
                                    case 1: {
                                            if (CodeList[i].End.x != CodeList[i].Start.x || CodeList[i].End.y != CodeList[i].Start.y) {
                                                tempout = "G1";
                                                if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                                if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                                if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                                if (CodeList[i].G53present) tempout = tempout + " G53";
                                                GCodeOutputFile->Add(tempout);
                                                endZ = CodeList[i].End.y;
                                                }
                                            }
                                            break;
                                    case 2: {
                                            tempout = "G3";
                                            if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                            tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                            tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                            tempout = tempout + " I" + DP(4,CodeList[i].Centre.i);
                                            //tempout = tempout + " J" + DP(4,CodeList[i].Centre.j);
                                            //tempout = tempout + " K" + DP(4,CodeList[i].Centre.k);
                                            tempout = tempout + " K" + DP(4,CodeList[i].Centre.j);
                                            if (CodeList[i].G53present) tempout = tempout + " G53";
                                            GCodeOutputFile->Add(tempout);
                                            endZ = CodeList[i].End.y;
                                            }
                                            break;
                                    case 3: {
                                            tempout = "G2";
                                            if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                            tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                            tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                            tempout = tempout + " I" + DP(4,CodeList[i].Centre.i);
                                            //tempout = tempout + " J" + DP(4,CodeList[i].Centre.j);
                                            //tempout = tempout + " K" + DP(4,CodeList[i].Centre.k);
                                            tempout = tempout + " K" + DP(4,CodeList[i].Centre.j);
                                            if (CodeList[i].G53present) tempout = tempout + " G53";
                                            GCodeOutputFile->Add(tempout);
                                            endZ = CodeList[i].End.y;
                                            }
                                            break;
                                    case 4: {
                                            tempout = "G4";
                                            if (CodeList[i].Start.x != 0.0) tempout = tempout + " P" + DP(4,CodeList[i].Start.x);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 50: {
                                            tempout = "G50";
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 51: {
                                            tempout = "G51";
                                            if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                            if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 92:
                                    case 52: {
                                            tempout = "G" + AnsiString(CodeList[i].Codeno);
                                            if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].End.x);
                                            if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].End.y);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 61: {
                                            tempout = "G61";
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 64: {
                                            tempout = "G64";
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    default:    break;
                                    }
                                }  //end of G codes
                            if (CodeList[i].Instruction == "M") {
                                switch ((int) CodeList[i].Codeno) {
                                    /*  don't want to put any stops in the output
                                    case 60:
                                    case 47:
                                    case 30:
                                    case 2:
                                    case 1:
                                    case 0: {
                                            tempout = "M" + AnsiString(CodeList[i].Codeno);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                            */
                                    case 5:
                                    case 4:
                                    case 3: {
                                            tempout = "M" + AnsiString(CodeList[i].Codeno);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 9:
                                    case 8:
                                    case 7: {
                                            tempout = "M" + AnsiString(CodeList[i].Codeno);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    case 49:
                                    case 48: {
                                            tempout = "M" + AnsiString(CodeList[i].Codeno);
                                            GCodeOutputFile->Add(tempout);
                                            }
                                            break;
                                    default:    break;
                                    }
                                }  //end of M codes
                            //don't bother with N codes
                            if (CodeList[i].Instruction == "P") {
                                GCodeOutputFile->Add("P" + AnsiString(CodeList[i].Codeno));
                                } //End of P code
                            if (CodeList[i].Instruction == "R") {
                                GCodeOutputFile->Add("R" + AnsiString(CodeList[i].Codeno));
                                } //End of R code
                            if (CodeList[i].Instruction == "S") {
                                GCodeOutputFile->Add("S" + AnsiString(CodeList[i].Codeno));
                                } //End of S code
                            if (CodeList[i].Instruction == "T") {
                                GCodeOutputFile->Add("T" + AnsiString(CodeList[i].Codeno));
                                } //End of T code
                            }  //end of 'if (CodeList[i].Comment) ... else...'
                        }  //end of 'for (int i = 0; i < CodeList.Length; i++)'
                    //forward run complete, go to clearance height
                    GCodeOutputFile->Add("(Go to clearance plane)");
                    GCodeOutputFile->Add("G0 Z" + DP(4,ClearancePlane));

                    //run through codelist backwards
                    currentangle += StepAngle;
                    if (currentangle < RotationAngle) {
                        GCodeOutputFile->Add("(Reverse Pass at angle " + DP(4,currentangle) + " )");
                        GCodeOutputFile->Add("G0 A" + DP(4,currentangle));
                        GCodeOutputFile->Add("G1 Z" + DP(4,endZ));
                        //now to step backwards through the codelist
                        for (int i = CodeList.Length - 1; i > -1 ; i--) {
                            if (CodeList[i].Comment) {
                                GCodeOutputFile->Add(CodeList[i].Instruction);
                                } else {
                                if (CodeList[i].Instruction == "F") {
                                    GCodeOutputFile->Add("F" + DP(4,CodeList[i].Start.feedrate));
                                    } //End of F code
                                if (CodeList[i].Instruction == "G") {
                                    switch ((int) CodeList[i].Codeno) {
                                        case 0: {
                                                if (CodeList[i].End.x != CodeList[i].Start.x || CodeList[i].End.y != CodeList[i].Start.y) {
                                                    tempout = "G0";
                                                    if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                    if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                    GCodeOutputFile->Add(tempout);
                                                    }
                                                }
                                                break;
                                        case 1: {
                                                if (CodeList[i].End.x != CodeList[i].Start.x || CodeList[i].End.y != CodeList[i].Start.y) {
                                                    tempout = "G1";
                                                    if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                                    if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                    if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                    if (CodeList[i].G53present) tempout = tempout + " G53";
                                                    GCodeOutputFile->Add(tempout);
                                                    if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) {
                                                        tempout = "F" + DP(4,CodeList[i].Start.feedrate);
                                                        GCodeOutputFile->Add(tempout);
                                                        }
                                                    }
                                                }
                                                break;
                                        case 2: {
                                                tempout = "G2";
                                                if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                                tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                tempout = tempout + " I" + DP(4,CodeList[i].Centre.i);
                                                //tempout = tempout + " J" + DP(4,CodeList[i].Centre.j);
                                                //tempout = tempout + " K" + DP(4,CodeList[i].Centre.k);
                                                tempout = tempout + " K" + DP(4,CodeList[i].Centre.j);
                                                if (CodeList[i].G53present) tempout = tempout + " G53";
                                                GCodeOutputFile->Add(tempout);
                                                if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) {
                                                    tempout = "F" + DP(4,CodeList[i].Start.feedrate);
                                                    GCodeOutputFile->Add(tempout);
                                                    }
                                                }
                                                break;
                                        case 3: {
                                                tempout = "G3";
                                                if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) tempout = tempout + " F" + DP(4,CodeList[i].End.feedrate);
                                                tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                tempout = tempout + " I" + DP(4,CodeList[i].Centre.i);
                                                //tempout = tempout + " J" + DP(4,CodeList[i].Centre.j);
                                                //tempout = tempout + " K" + DP(4,CodeList[i].Centre.k);
                                                tempout = tempout + " K" + DP(4,CodeList[i].Centre.j);
                                                if (CodeList[i].G53present) tempout = tempout + " G53";
                                                GCodeOutputFile->Add(tempout);
                                                if (CodeList[i].End.feedrate != CodeList[i].Start.feedrate) {
                                                    tempout = "F" + DP(4,CodeList[i].Start.feedrate);
                                                    GCodeOutputFile->Add(tempout);
                                                    }
                                                }
                                                break;
                                        case 4: {
                                                tempout = "G4";
                                                if (CodeList[i].Start.x != 0.0) tempout = tempout + " P" + DP(4,CodeList[i].Start.x);
                                                GCodeOutputFile->Add(tempout);
                                                }
                                                break;
                                        case 51:
                                        case 50: {
                                                if (CodeList[i].Start.x == 1.0 && CodeList[i].Start.y == 1.0) {
                                                    tempout = "G50";
                                                    GCodeOutputFile->Add(tempout);
                                                    } else {
                                                    tempout = "G51";
                                                    if (CodeList[i].Start.x != 1.0) tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                    if (CodeList[i].Start.y != 1.0) tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                    GCodeOutputFile->Add(tempout);
                                                    }
                                                }
                                                break;
                                        case 92:
                                        case 52: {
                                                tempout = "G" + AnsiString(CodeList[i].Codeno);
                                                if (CodeList[i].End.x != CodeList[i].Start.x) tempout = tempout + " X" + DP(4,CodeList[i].Start.x);
                                                if (CodeList[i].End.y != CodeList[i].Start.y) tempout = tempout + " Z" + DP(4,CodeList[i].Start.y);
                                                GCodeOutputFile->Add(tempout);
                                                }
                                                break;
                                        case 64:
                                        case 61: {
                                                if ((TPathControl) CodeList[i].Start.x == TExact) {
                                                    GCodeOutputFile->Add("G61");
                                                    } else {
                                                    GCodeOutputFile->Add("G64");
                                                    }
                                                }
                                                break;
                                        default:    break;
                                        }
                                    }  //end of G codes
                                if (CodeList[i].Instruction == "M") {
                                    switch ((int) CodeList[i].Codeno) {
                                        /*   get rid of all stops
                                        case 60:
                                        case 47:
                                        case 30:
                                        case 2:
                                        case 1:
                                        case 0: {
                                                tempout = "M" + AnsiString(CodeList[i].Codeno);
                                                GCodeOutputFile->Add(tempout);
                                                }
                                                break;
                                                */
                                        case 5:
                                        case 4:
                                        case 3: {
                                                switch ((TSpindleDir) CodeList[i].Start.x) {
                                                    case TCW:   {
                                                                GCodeOutputFile->Add("M3");
                                                                }
                                                                break;
                                                    case TCCW:   {
                                                                GCodeOutputFile->Add("M4");
                                                                }
                                                                break;
                                                    case TStop: {
                                                                GCodeOutputFile->Add("M5");
                                                                }
                                                                break;
                                                    default:    break;
                                                    }
                                                }
                                                break;
                                        case 9:
                                        case 8:
                                        case 7: {
                                                switch ((TCoolant) CodeList[i].Start.x) {
                                                    case TMist: {
                                                                GCodeOutputFile->Add("M7");
                                                                }
                                                                break;
                                                    case TFlood: {
                                                                GCodeOutputFile->Add("M8");
                                                                }
                                                                break;
                                                    case TNone: {
                                                                GCodeOutputFile->Add("M9");
                                                                }
                                                                break;
                                                    default:    break;
                                                    }
                                                }
                                                break;
                                        case 49:
                                        case 48: {
                                                tempout = "M" + AnsiString(CodeList[i].Codeno);
                                                GCodeOutputFile->Add(tempout);
                                                }
                                                break;
                                        default:    break;
                                        }
                                    }  //end of M codes
                                //don't bother with N codes
                                if (CodeList[i].Instruction == "P") {
                                    GCodeOutputFile->Add("P" + AnsiString(CodeList[i].Codeno));
                                    } //End of P code
                                if (CodeList[i].Instruction == "R") {
                                    GCodeOutputFile->Add("R" + AnsiString(CodeList[i].Start.x));
                                    } //End of R code
                                if (CodeList[i].Instruction == "S") {
                                    GCodeOutputFile->Add("S" + AnsiString(CodeList[i].Start.x));
                                    } //End of S code
                                if (CodeList[i].Instruction == "T") {
                                    GCodeOutputFile->Add("T" + AnsiString(CodeList[i].Start.x));
                                    } //End of T code
                                }  //end of 'if (CodeList[i].Comment) ... else...'
                            }  //end of 'for (int i = CodeList.Length - 1; i > -1 ; i--)'


                        GCodeOutputFile->Add("(Go to clearance plane)");
                        GCodeOutputFile->Add("G0 Z" + DP(4,ClearancePlane));

                        //end of running through codelist backwards
                        currentangle += StepAngle;
                        } //end of 'if (currentangle < RotationAngle)'
                    } //end of ' while (currentangle < RotationAngle)'
                } else {
                Errors.Length++;
                Errors[Errors.Length - 1] = TNoAction;
                }//end of 'if (counter < CodeList.Length)... else...'



            //calculate how long the machining program takes
            /*
            tempout = "";
            double hours = (double) ((int)(currentTime/3600.0));
            currentTime -= hours * 3600.0;
            double minutes = (double) ((int)(currentTime/60.0));
            currentTime -= minutes * 60.0;

            if (hours > 0.0) {
                tempout = tempout + AnsiString(hours) + " Hours, ";
                }
            if (minutes > 0.0) {
                tempout = tempout + AnsiString(minutes) + " Minutes, ";
                }
            tempout = tempout + DP(2,currentTime) + " Seconds.";


            tempout = "(Total Running Time minimum = " + tempout + " )";
            GCodeOutputFile->Insert(5,tempout);
            */

            //Add footer
            GCodeOutputFile->Add("(Footer)");
            GCodeOutputFile->Add(Footer->Text);

            //Finish by saving then clearing memory
            Label3->Caption = "Saving File";
            Refresh();
            /*
            //debug
            for (int i = 0; i < GCodeInputFile->Count; i++) {
                GCodeOutputFile->Add("(" + GCodeInputFile->Strings[i] + ")");
                }
            */

            GCodeOutputFile->SaveToFile(Edit2->Text);

            Label3->Caption = "Clearing Memory";
            Refresh();

            delete GCodeInputFile;
            delete GCodeOutputFile;

            if (!unknowncode) {
                Label3->Caption = "Finished";
                } else {
                Label3->Caption = "Unknown Codes encountered, please check file";
                AnsiString Error = Errors.Length > 1 ? "s were " : " was ";
                Error = "The following Error" + Error + "encountered:\n";
                for (int i = 1; i < Errors.Length; i++) {
                    switch (Errors[i]) {
                        case TXZPresent  :  {
                                            Error = Error + "\tG18: XZ Plane definition found.\n";
                                            }
                                            break;
                        case TYZPresent  :  {
                                            Error = Error + "\tG19: YZ Plane definition encountered.\n";
                                            }
                                            break;
                        case TZeroSizeArc:  {
                                            Error = Error + "\tZero size arc met.\n";
                                            }
                                            break;
                        case TPlaneType  :  {
                                            Error = Error + "\tUnknown Working Plane definition specified. Possible program internal error.\n";
                                            }
                                            break;
                        case TUnknownMCode: {
                                            Error = Error + "\tUnknown M Code Encountered.\n";
                                            }
                                            break;
                        case TUnknownGCode: {
                                            Error = Error + "\tUnknown G Code Encountered.\n";
                                            }
                                            break;
                        case TAAxisInstruction: {
                                            Error = Error + "\tThe input file contains A Axis Instructions.\n";
                                            }
                                            break;
                        case TBAxisInstruction: {
                                            Error = Error + "\tThe input file contains B Axis Instructions.\n";
                                            }
                                            break;
                        case TCAxisInstruction: {
                                            Error = Error + "\tThe input file contains C Axis Instructions.\n";
                                            }
                                            break;
                        case TCoOrdRotation: {
                                            Error = Error + "\tThe input file contains Co-Ordinate rotations.\n";
                                            }
                                            break;
                        case TNoAction:     {
                                            Error = Error + "\tThe input file contains no movement codes (G0, G1, G2 or G3).\n";
                                            }
                                            break;
                        default:            {
                                            Error = Error + "\tProgram internal error.\n";
                                            }
                                            break;
                        }
                    }
                Errors.Length = 0;
                }
            //delete Errors;
            CodeList.Length = 0;
            //delete CodeList;
            Refresh();
            }  //end of try  statement

        __finally {
            ProfileButton->Enabled = true;
            OriginalGCodeButton->Enabled = true;
            NewGCodeButton->Enabled = true;
            HelpButton->Enabled = true;
            CloseButton->Enabled = true;
            Edit1->Enabled = true;
            Edit2->Enabled = true;
            Edit3->Enabled = true;
            Edit4->Enabled = true;
            Edit5->Enabled = true;
            Header->Enabled = true;
            Footer->Enabled = true;
            Refresh();
            }

        } else {
        if (FileExists(Edit1->Text)) {
            Application->MessageBox("An existing converted file will get overwritten.\n\nPlease delete file or change converted file name","Warning", MB_OK | MB_ICONERROR);
            } else {
            Application->MessageBox("The input file does not exist.\n\nPlease check","Warning", MB_OK | MB_ICONERROR);
            }
        Label3->Caption = "Conversion Aborted";
        }
    
}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::HelpButtonClick(TObject *Sender)
{
    //Application->MessageBox("This program takes an XY plane profile in the +Y half of the co-ordinate system, and wraps the code around the A axis at intervals specified by\nthe 'Angular Step' for a total angle specified by 'Total Angle'.\n\nTo operate, select 2 file names, angle and angular step,\nthen fill in the pertinant header/footer data and press the\n'Profile' button.\n","Help on G Code Profile Development Program",MB_OK | MB_ICONINFORMATION);
    AnsiString Address = ExtractFilePath(Application->ExeName) + "\\G Code Profile Help.pdf";
    ShellExecute(Application->Handle,NULL,Address.c_str(),NULL,NULL,0);
    Address.~AnsiString();

    
}
//---------------------------------------------------------------------------
void __fastcall TProfileForm::CloseButtonClick(TObject *Sender)
{
    Close();    
}
//---------------------------------------------------------------------------

